/*
 *  ircd-hybrid: an advanced Internet Relay Chat Daemon(ircd).
 *  event.c: Event functions.
 *
 *  Copyright (C) 1998-2000 Regents of the University of California
 *  Copyright (C) 2001-2002 Hybrid Development Team
 *
 *  Code borrowed from the squid web cache by Adrian Chadd.
 *  Original header:
 *
 *  DEBUG: section 41   Event Processing
 *  AUTHOR: Henrik Nordstrom
 *
 *  SQUID Internet Object Cache  http://squid.nlanr.net/Squid/
 *  ----------------------------------------------------------
 *
 *  Squid is the result of efforts by numerous individuals from the
 *  Internet community.  Development is led by Duane Wessels of the
 *  National Laboratory for Applied Network Research and funded by the
 *  National Science Foundation.  Squid is Copyrighted (C) 1998 by
 *  the Regents of the University of California.  Please see the
 *  COPYRIGHT file for full details.  Squid incorporates software
 *  developed and/or copyrighted by other sources.  Please see the
 *  CREDITS file for full details.
 */

/*
 * How its used:
 *
 * Should be pretty self-explanatory. Events are added to the static
 * array event_table with a frequency time telling eventRun how often
 * to execute it.
 */

#include <stdlib.h>
#include <sys/types.h>
#include <time.h>

#include "event.h"
#include "numeric.h"
#include "struct.h"
#include "proto.h"

static const char *last_event_ran = NULL;
struct ev_entry event_table[MAX_EVENTS];
static int event_count = 0;
static time_t event_time_min = -1;


/*
 * void eventAdd(const char *name, EVH *func, void *arg, time_t when)
 *
 * Input: Name of event, function to call, arguments to pass, and frequency
 *	  of the event.
 * Output: None
 * Side Effects: Adds the event to the event list.
 */

void
eventAdd(const char *name, EVH *func, void *arg, time_t when)
{
  int i;
  
  /* find first inactive index, or use next index */
  for (i = 0; i < event_count; i++)
    if (!event_table[i].active)
      break;

  if (i >= event_count)
    event_count = i + 1;

  event_table[i].func = func;
  event_table[i].name = name;
  event_table[i].arg = arg;
  event_table[i].when = time(NULL) + when;
  event_table[i].frequency = when; 
  event_table[i].active = 1;

  if ((event_table[i].when < event_time_min) || (event_time_min == -1))
    event_time_min = event_table[i].when;
}

/*
 * void eventDelete(EVH *func, void *arg)
 *
 * Input: Function handler, argument that was passed.
 * Output: None
 * Side Effects: Removes the event from the event list
 */

void
eventDelete(EVH *func, void *arg)
{
  int i;
 
  i = eventFind(func, arg);

  if (i == -1)
    return;
  
  event_table[i].name = NULL;
  event_table[i].func = NULL;
  event_table[i].arg = NULL;
  event_table[i].active = 0;
}

/* 
 * void eventAddIsh(const char *name, EVH *func, void *arg, time_t delta_isa)
 *
 * Input: Name of event, function to call, arguments to pass, and frequency
 *	  of the event.
 * Output: None
 * Side Effects: Adds the event to the event list within +- 1/3 of the
 *	         specified frequency.
 */
 
void
eventAddIsh(const char *name, EVH *func, void *arg, time_t delta_ish)
{
  if (delta_ish >= 3.0)
    {
      const time_t two_third = (2 * delta_ish) / 3;
      delta_ish = two_third + ((random() % 1000) * two_third) / 1000;
      /*
       * XXX I hate the above magic, I don't even know if its right.
       * Grr. -- adrian
       */
    }
  eventAdd(name, func, arg, delta_ish);
}

/*
 * void eventRun(void)
 *
 * Input: None
 * Output: None
 * Side Effects: Runs pending events in the event list
 */

void
eventRun(void)
{
  int i;
  time_t current_time;

  if (event_count == 0)
    return;

  current_time = time(NULL);

  for (i = 0; i < event_count; i++)
    {
      if (event_table[i].active && (event_table[i].when <= current_time))
        {
          last_event_ran = event_table[i].name;
          event_table[i].func(event_table[i].arg);
          event_table[i].when = current_time + event_table[i].frequency;
          event_time_min = -1;
        }
    }
}


/*
 * time_t eventNextTime(void)
 * 
 * Input: None
 * Output: Specifies the next time eventRun() should be run
 * Side Effects: None
 */
 
time_t
eventNextTime(void)
{
  int i;

  if (event_count == 0)
    return (time(NULL)+1);
  else if (event_time_min == -1)
    {
      for (i = 0; i < event_count; i++)
        {
          if (event_table[i].active && ((event_table[i].when < event_time_min) || (event_time_min == -1)))
            event_time_min = event_table[i].when;
        }
    }
  return event_time_min;
}

/*
 * void eventInit(void)
 *
 * Input: None
 * Output: None
 * Side Effects: Initializes the event system. 
 */
void
eventInit(void)
{
  last_event_ran = NULL;
  event_count = 0;
}

/*
 * int eventFind(EVH *func, void *arg)
 *
 * Input: Event function and the argument passed to it
 * Output: Index to the slow in the event_table
 * Side Effects: None
 */

int
eventFind(EVH *func, void *arg)
{
  int i;
  for (i = 0; i < event_count; i++)
    {
      if ((event_table[i].func == func) &&
          (event_table[i].arg == arg) &&
          event_table[i].active)
        return i;
    }
  return -1;
}

/* 
 * void show_events(char *origin)
 *
 * Input: Client requesting the event
 * Output: List of events
 * Side Effects: None
 */

void
show_events(char *origin)
{
  int i;
  time_t current_time = time(NULL);

  if (last_event_ran)
    sts(":%s %d %s :*** Last event to run: %s", settings.name, RPL_STATSDEBUG,
	origin, last_event_ran);

  sts(":%s %d %s :*** Operation            Next Execution", settings.name,
	RPL_STATSDEBUG, origin);

  for (i = 0; i < event_count; i++)
    {
      if (event_table[i].active)
        {
          sts(":%s %d %s :*** %-20s %-3d seconds", settings.name,
		RPL_STATSDEBUG, origin, event_table[i].name,
		(int)(event_table[i].when - current_time));
        }
    }
  sts(":%s %d %s :*** Finished", settings.name, RPL_STATSDEBUG, origin);
}

/* 
 * void set_back_events(time_t by)
 * Input: Time to set back events by.
 * Output: None.
 * Side-effects: Sets back all events by "by" seconds.
 */
void
set_back_events(time_t by)
{
  int i;
  for (i = 0; i < event_count; i++)
    if (event_table[i].when > by)
      event_table[i].when -= by;
    else
      event_table[i].when = 0;
}
